# supported devices
#   Mi Body Composition Scale 2 (XMTZC05HM)

import logging

from datetime import date, datetime
from struct import unpack
from typing import Any, Dict, List, Union

import miblepy.devices.xbm as xbm

from bluepy.btle import BTLEDisconnectError, BTLEManagementError, DefaultDelegate, ScanEntry, Scanner
from miblepy import ATTRS
from miblepy.deviceplugin import MibleDevicePlugin


PLUGIN_NAME = "BodyCompScale"

SCAN_TIMEOUT = 10
UNITS = {2: "kg", 3: "lbs"}
DATA_KEYS = (
    "unit_id",
    "control",
    "year",
    "month",
    "day",
    "hour",
    "min",
    "sec",
    "impedance",
    "weight",
)


class BodyCompScale(MibleDevicePlugin, DefaultDelegate):

    plugin_id = "bodycompscale"
    plugin_name = "BodyCompScale"
    plugin_description = "suports the Mi Body Composition Scale 2 (XMTZC05HM) / Xiaomi Scale 2 (XMTZC02HM)"

    def __init__(self, mac: str, interface: str, **kwargs: Any):
        self.users: List[Dict[str, Union[str, int, float, date]]] = kwargs.get("users", [])
        self.scanner: Scanner = None
        self.data: Dict[str, Any] = {}

        super().__init__(mac, interface, **kwargs)

    def fetch_data(self, **kwargs: Any) -> Dict[str, Any]:
        """Get data from device."""

        # attach notification handler
        self.scanner = Scanner(iface=int(self.interface.replace("hci", ""))).withDelegate(self)

        try:
            self.scanner.scan(SCAN_TIMEOUT)
        except BTLEDisconnectError as error:
            logging.error(f"btle disconnected: {error}")
        except BTLEManagementError as error:
            logging.error(f"(temporary) bluetooth connection error: {error}")

        return self.data

    def get_age(self, birthdate: Any) -> int:
        today = date.today()
        return int(today.year - birthdate.year - ((today.month, today.day) < (birthdate.month, birthdate.day)))

    def find_user(self, weight: float) -> Dict[str, Any]:

        current_user: Dict[str, Any] = {}

        # determine current user by weight
        for user in self.users:

            if not all([user["weightOver"], user["weightBelow"]]):
                continue

            if user["weightOver"] < weight < user["weightBelow"]:  # type: ignore
                # current user found, fill profile values and exit
                current_user = user
                current_user[ATTRS.WEIGHT.value] = weight
                current_user[ATTRS.AGE.value] = self.get_age(current_user["birthdate"])
                break

        return current_user

    def handleDiscovery(self, dev: ScanEntry, new_dev: bool, new_data: bool) -> None:

        if not dev.addr == self.mac.lower() or not new_dev or not new_data:
            return

        for (sdid, _, data) in dev.getScanData():

            # Mi Body Composition Scale 2 (XMTZC05HM) / Xiaomi Scale 2 (XMTZC02HM)
            if not data.startswith("1b18") or sdid != 22:
                continue

            # 15b in little endian
            #   0-1: identifier?
            #     2: unit
            #     3: control byte
            #   4-5: year
            #     6: month
            #     7: day
            #     8: hour
            #     9: min
            #     10: sec
            #  11-12: impedance
            #  13-14: weight

            # unpack bytes to dictionary
            measured = dict(zip(DATA_KEYS, unpack("<xxBBHBBBBBHH", bytes.fromhex(data))))

            # check if we got a proper measurement
            measurement_stabilized = measured["control"] & (1 << 5)
            impedance_available = measured["control"] & (1 << 1)

            # pick unit
            unit = UNITS.get(measured["unit_id"], None)
            # calc weight based on unit
            weight = measured["weight"] / 100 / 2 if measured["unit_id"] == 2 else measured["weight"] / 100

            # check if we got a proper measurement
            if not all([measurement_stabilized, unit]):
                logging.debug(f"missing data! weight: {weight} | unit: {unit} | impedance: {measured['impedance']}")
                continue

            # create datetime
            measurement_datetime = datetime(
                measured["year"], measured["month"], measured["day"], measured["hour"], measured["min"], measured["sec"]
            )

            # find the current user based on its weight
            if user := self.find_user(weight):

                bm = xbm.BodyMetrics(
                    user[ATTRS.WEIGHT.value],
                    user[ATTRS.HEIGHT.value],
                    user[ATTRS.AGE.value],
                    user[ATTRS.SEX.value],
                    measured["impedance"],
                )

                attributes = {
                    ATTRS.USER.value: user[ATTRS.USER.value],
                    ATTRS.AGE.value: user[ATTRS.AGE.value],
                    ATTRS.SEX.value: user[ATTRS.SEX.value],
                    ATTRS.HEIGHT.value: user[ATTRS.HEIGHT.value],
                    ATTRS.WEIGHT.value: f"{weight:.2f}",
                    ATTRS.UNIT.value: unit,
                    ATTRS.BASAL_METABOLISM.value: f"{bm.get_bmr():.2f}",
                    ATTRS.VISCERAL_FAT.value: f"{bm.getVisceralFat():.2f}",
                    ATTRS.BMI.value: f"{bm.getBMI():.2f}",
                    ATTRS.TIMESTAMP.value: measurement_datetime.isoformat(),
                }

                # if we got a valid impedance, we can add more metrics
                if impedance_available:
                    attributes.update(
                        {
                            ATTRS.WATER.value: f"{bm.getWaterPercentage():.2f}",
                            ATTRS.BONE_MASS.value: f"{bm.getBoneMass():.2f}",
                            ATTRS.BODY_FAT.value: f"{bm.getFatPercentage():.2f}",
                            ATTRS.LEAN_BODY_MASS.value: f"{bm.get_lbm_coefficient():.2f}",
                            ATTRS.MUSCLE_MASS.value: f"{bm.getMuscleMass():.2f}",
                            ATTRS.PROTEIN.value: f"{bm.getProteinPercentage():.2f}",
                        }
                    )

                self.data.update(
                    {
                        "name": PLUGIN_NAME,
                        "sensors": [
                            {
                                "name": f"{self.alias} {user[ATTRS.USER.value]}",
                                "value_template": "{{value_json." + ATTRS.WEIGHT.value + "}}",
                                "entity_type": ATTRS.WEIGHT,
                                "own_state_topic": True,
                            },
                        ],
                        "attributes": attributes,
                    }
                )



from abc import ABC, abstractmethod
from typing import Any, Dict


class MibleDevicePlugin(ABC):
    def __init__(self, mac: str, interface: str, **kwargs: Any):
        self.mac = mac
        self.interface = interface

        if alias := kwargs.get("alias"):
            self.alias = alias
        else:
            self.alias = mac

        super().__init__()

    @property
    @abstractmethod
    def plugin_id(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def plugin_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def plugin_description(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def fetch_data(self, **kwargs: Any) -> Dict[str, Any]:
        raise NotImplementedError



from datetime import datetime
from typing import Any, Dict

from bluepy.btle import Peripheral
from miblepy import ATTRS
from miblepy.deviceplugin import MibleDevicePlugin


class FlowerCare(MibleDevicePlugin):

    plugin_id = "flowercare"
    plugin_name = "FlowerCare"
    plugin_description = "suports the VegTrug/Xiaomi/Mi Flora plant sensors"

    def __init__(self, mac: str, interface: str, **kwargs: Any):
        super().__init__(mac, interface, **kwargs)

    def fetch_data(self, **kwargs: Any) -> Dict[str, Any]:
        """Get data from one Sensor."""

        # connect to device
        peripheral = Peripheral(self.mac, iface=int(self.interface.replace("hci", "")))

        # enable reading of values
        peripheral.writeCharacteristic(0x33, bytes([0xA0, 0x1F]), withResponse=True)

        # 7b in little endian
        #    0: battery level
        #    1: unknown
        #  2-6: firmware version
        battery_and_firmware: bytes = peripheral.readCharacteristic(0x38)
        battery_level = int.from_bytes(battery_and_firmware[:1], byteorder="little")
        firmware_version = str(battery_and_firmware[2:].decode("utf-8"))

        # 16b in little endian
        #   0-1: temperature in 0.1 °C
        #     2: unknown
        #   3-6: brightness in lux
        #     7: moisture in %
        #   8-9: conductivity in µS/cm
        # 10-15: unknown
        data: bytes = peripheral.readCharacteristic(0x35)

        plugin_data: Dict[str, Any] = {
            "name": self.plugin_name,
            "sensors": [
                {
                    "name": f"{self.alias} {ATTRS.TEMPERATURE.value.capitalize()}",
                    "value_template": "{{value_json." + ATTRS.TEMPERATURE.value + "}}",
                    "entity_type": ATTRS.TEMPERATURE,
                },
                {
                    "name": f"{self.alias} {ATTRS.BRIGHTNESS.value.capitalize()}",
                    "value_template": "{{value_json." + ATTRS.BRIGHTNESS.value + "}}",
                    "entity_type": ATTRS.BRIGHTNESS,
                },
                {
                    "name": f"{self.alias} {ATTRS.MOISTURE.value.capitalize()}",
                    "value_template": "{{value_json." + ATTRS.MOISTURE.value + "}}",
                    "entity_type": ATTRS.MOISTURE,
                },
                {
                    "name": f"{self.alias} {ATTRS.CONDUCTIVITY.value.capitalize()}",
                    "value_template": "{{value_json." + ATTRS.CONDUCTIVITY.value + "}}",
                    "entity_type": ATTRS.CONDUCTIVITY,
                },
                {
                    "name": f"{self.alias} {ATTRS.BATTERY.value.capitalize()}",
                    "value_template": "{{value_json." + ATTRS.BATTERY.value + "}}",
                    "entity_type": ATTRS.BATTERY,
                },
            ],
            "attributes": {
                ATTRS.BATTERY.value: str(battery_level),
                ATTRS.TEMPERATURE.value: str(int.from_bytes(data[0:2], byteorder="little") / 10),
                ATTRS.BRIGHTNESS.value: str(int.from_bytes(data[3:6], byteorder="little")),
                ATTRS.MOISTURE.value: str(int.from_bytes(data[7:8], byteorder="little")),
                ATTRS.CONDUCTIVITY.value: str(int.from_bytes(data[8:10], byteorder="little")),
                ATTRS.FW_VERSION.value: firmware_version,
                ATTRS.TIMESTAMP.value: datetime.now().isoformat(),
            },
        }

        return plugin_data



from datetime import datetime
from typing import Any, Dict

from bluepy.btle import DefaultDelegate, Peripheral
from miblepy import ATTRS
from miblepy.deviceplugin import MibleDevicePlugin


class LYWSD03MMC(MibleDevicePlugin, DefaultDelegate):

    plugin_id = "lywsd03mmc"
    plugin_name = "LYWSD03MMC"
    plugin_description = "suports the Temperature/Humidity LCD BLE sensor LYWSD03MMC from Mi/Xiaomi"

    def __init__(self, mac: str, interface: str, **kwargs: Any):
        self.peripheral: Peripheral = None
        self.data: Dict[str, Any] = {}

        super().__init__(mac, interface, **kwargs)

    def fetch_data(self, **kwargs: Any) -> Dict[str, Any]:
        # connect to device
        self.peripheral = Peripheral(self.mac, iface=int(self.interface.replace("hci", "")))

        # attach notification handler
        self.peripheral.setDelegate(self)

        # safe power: https://github.com/JsBergbau/MiTemperature2/issues/18#issuecomment-590986874
        self.peripheral.writeCharacteristic(0x46, bytes([0xF4, 0x01, 0x00]), withResponse=True)

        if self.peripheral.waitForNotifications(10000):
            self.peripheral.disconnect()

        return self.data

    def handleNotification(self, cHandle: int, data: bytes) -> None:
        if cHandle != 0x36:
            return

        # parse data
        voltage = int.from_bytes(data[3:5], byteorder="little") / 1000

        self.data.update(
            {
                "name": self.plugin_name,
                "sensors": [
                    {
                        "name": f"{self.alias} {ATTRS.TEMPERATURE.value.capitalize()}",
                        "value_template": "{{value_json." + ATTRS.TEMPERATURE.value + "}}",
                        "entity_type": ATTRS.TEMPERATURE,
                    },
                    {
                        "name": f"{self.alias} {ATTRS.HUMIDITY.value.capitalize()}",
                        "value_template": "{{value_json." + ATTRS.HUMIDITY.value + "}}",
                        "entity_type": ATTRS.HUMIDITY,
                    },
                ],
                "attributes": {
                    # 3.1 or above --> 100% 2.1 --> 0 %
                    ATTRS.BATTERY.value: min(int(round((voltage - 2.1), 2) * 100), 100),
                    ATTRS.VOLTAGE.value: str(voltage),
                    ATTRS.TEMPERATURE.value: str(int.from_bytes(data[0:2], byteorder="little", signed=True) / 100),
                    ATTRS.HUMIDITY.value: str(int.from_bytes(data[2:3], byteorder="little")),
                    ATTRS.TIMESTAMP.value: str(datetime.now().isoformat()),
                },
            }
        )

        self.peripheral.disconnect()



# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os.path

readme = ''
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, 'README.rst')
if os.path.exists(readme_path):
    with open(readme_path, 'rb') as stream:
        readme = stream.read().decode('utf8')

setup(
    long_description=readme,
    name='miblepy',
    version='0.4.6',
    description='miblepy fetches data from various (Xiaomi/Mijia/Mi) Bluetooth LE devices and push it to a MQTT broker in a coordinated, sequential manner.',
    python_requires='==3.*,>=3.8.0',
    project_urls={
        "homepage": "https://github.com/benleb/miblepy",
        "repository": "https://github.com/benleb/miblepy"
    },
    author='Ben Lebherz',
    author_email='git@benleb.de',
    license='MIT',
    keywords='bluetooth smart-home sensor xiaomi home-assistant',
    classifiers=[
        'Development Status :: 4 - Beta', 'Environment :: Console',
        'Intended Audience :: Developers', 'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8', 'Topic :: Home Automation'
    ],
    entry_points={"console_scripts": ["mible = miblepy.mible.cli:cli"]},
    packages=[
        'miblepy', 'miblepy.devices', 'miblepy.devices.xbm', 'miblepy.mible'
    ],
    package_dir={"": "."},
    package_data={},
    install_requires=[
        'bluepy==1.*,>=1.3.0', 'click==7.*,>=7.1.2', 'paho-mqtt==1.*,>=1.5.0',
        'tomlkit==0.*,>=0.6.0'
    ],
    extras_require={
        "dev": [
            "bandit==1.*,>=1.6.2", "black==19.*,>=19.10.0.b0",
            "flake8==3.*,>=3.8.1", "isort==4.*,>=4.3.21", "mypy==0.*,>=0.770.0",
            "pycodestyle==2.*,>=2.6.0", "pytest==5.*,>=5.2.0"
        ]
    },
)
